import datetime
import typing
import uuid

from dataclasses_avroschema import fields

now = datetime.datetime.now()

PRIMITIVE_TYPES = (
    (str, fields.STRING),
    (int, fields.INT),
    (bool, fields.BOOLEAN),
    (float, fields.FLOAT),
    # (bytes, "bytes"),
)

PRIMITIVE_TYPES_AND_DEFAULTS = (
    (str, "test"),
    (int, 1),
    (bool, True),
    (float, 10.4),
    # (bytes, "test".encode()),
)

PRIMITIVE_TYPES_AND_INVALID_DEFAULTS = (
    (str, 1),
    (int, "test"),
    (bool, 10),
    (float, False),
    # (bytes, "test".encode()),
)

LIST_TYPE_AND_ITEMS_TYPE = (
    (str, "string"),
    (int, "int"),
    (bool, "boolean"),
    (float, "float"),
    (bytes, "bytes"),
)

LOGICAL_TYPES = (
    (datetime.date, fields.LOGICAL_DATE, now.date()),
    (datetime.time, fields.LOGICAL_TIME, now.time()),
    (datetime.datetime, fields.LOGICAL_DATETIME, now),
    (uuid.uuid4, fields.LOGICAL_UUID, uuid.uuid4()),
)

UNION_PRIMITIVE_ELEMENTS = (
    ((str, int), (fields.STRING, fields.INT)),
    ((str, None), (fields.STRING, fields.NULL)),
    (
        (datetime.date, datetime.datetime),
        (fields.PYTHON_TYPE_TO_AVRO[datetime.date], fields.PYTHON_TYPE_TO_AVRO[datetime.datetime],),
    ),
    ((float, str, int), (fields.FLOAT, fields.STRING, fields.INT)),
    ((str, float, int, bool), (fields.STRING, fields.FLOAT, fields.INT, fields.BOOLEAN),),
)


SEQUENCE_TYPES = (typing.List, typing.Sequence, typing.MutableSequence)
MAPPING_TYPES = (typing.Dict, typing.Mapping, typing.MutableMapping)

SEQUENCES_AND_TYPES = (
    (sequence, python_type, items_type) for sequence in SEQUENCE_TYPES for python_type, items_type in PRIMITIVE_TYPES
)

SEQUENCES_LOGICAL_TYPES = (
    (sequence, python_type, items_type, value)
    for sequence in SEQUENCE_TYPES
    for python_type, items_type, value in LOGICAL_TYPES
)

MAPPING_AND_TYPES = (
    (mapping, python_type, items_type) for mapping in MAPPING_TYPES for python_type, items_type in PRIMITIVE_TYPES
)

MAPPING_LOGICAL_TYPES = (
    (mapping, python_type, items_type, value)
    for mapping in MAPPING_TYPES
    for python_type, items_type, value in LOGICAL_TYPES
)

# Represent the logical types
# (python_type, avro_internal_type, logical_type)
LOGICAL_TYPES_AND_DEFAULTS = (
    (datetime.date, fields.INT, fields.DATE),
    (datetime.time, fields.INT, fields.TIME_MILLIS),
    (datetime.datetime, fields.LONG, fields.TIMESTAMP_MILLIS),
    (uuid.uuid4, fields.STRING, fields.UUID),
)

LOGICAL_TYPES_AND_INVALID_DEFAULTS = (
    (datetime.date, 1, None),
    (datetime.time, "test", None),
    (datetime.datetime, 10, None),
    (uuid.uuid4, 10, f"Invalid default type. Default should be {str} or {uuid.UUID}"),
)


class User:
    "User"
    first_name: str


avro_user = {
    "name": "User",
    "type": "record",
    "doc": "User",
    "fields": [{"name": "first_name", "type": "string"}],
}

ARRAY_WITH_UNION_TYPES = (
    (typing.Union[int, str], [fields.INT, fields.STRING], [10, 20, "test"]),
    (typing.Union[int, str, User], [fields.INT, fields.STRING, avro_user], [10, 20, "test"],),
)



import abc
import dataclasses
import inspect
import typing
from collections import OrderedDict

from dataclasses_avroschema import fields

try:
    import faust
except ImportError:  # pragma: no cover
    faust = None  # pragma: no cover


@dataclasses.dataclass
class BaseSchemaDefinition(abc.ABC):
    """
    Minimal Schema definition
    """

    type: str
    klass_or_instance: typing.Any

    @abc.abstractmethod
    def get_rendered_fields(self):
        ...  # pragma: no cover

    @abc.abstractmethod
    def render(self):
        ...  # pragma: no cover

    def get_schema_name(self):
        if inspect.isclass(self.klass_or_instance):
            return self.klass_or_instance.__name__
        return self.klass_or_instance.__class__.__name__

    def generate_documentation(self):
        doc = self.klass_or_instance.__doc__

        if doc is not None:
            return doc.replace("\n", "")

    @property
    def is_faust_record(self) -> bool:
        if faust:
            if inspect.isclass(self.klass_or_instance):
                return issubclass(self.klass_or_instance, faust.Record)
            return issubclass(self.klass_or_instance.__class__, faust.Record)

        return False


@dataclasses.dataclass
class AvroSchemaDefinition(BaseSchemaDefinition):
    aliases: typing.List[str] = None
    namespace: str = None
    fields: typing.List["fields.FieldType"] = None
    include_schema_doc: bool = True

    def __post_init__(self):
        self.generate_extra_avro_attributes()
        self.fields = self.parse_dataclasses_fields()

    def parse_dataclasses_fields(self) -> typing.List["fields.Field"]:
        if self.is_faust_record:
            return self.parse_faust_record_fields()
        return self.parse_fields()

    def parse_fields(self):
        return [
            fields.Field(
                dataclass_field.name,
                dataclass_field.type,
                dataclass_field.default,
                dataclass_field.default_factory,
                dataclass_field.metadata,
            )
            for dataclass_field in dataclasses.fields(self.klass_or_instance)
        ]

    def parse_faust_record_fields(self) -> typing.List["fields.Field"]:
        schema_fields = []

        for dataclass_field in dataclasses.fields(self.klass_or_instance):
            faust_field = dataclass_field.default

            if faust_field.required:
                default = dataclasses.MISSING
                default_factory = dataclasses.MISSING
            else:
                default = faust_field.default
                default_factory = dataclasses.MISSING

                if isinstance(default, dataclasses.Field):
                    default_factory = default.default_factory
                    default = dataclasses.MISSING

            schema_fields.append(fields.Field(dataclass_field.name, dataclass_field.type, default, default_factory))

        return schema_fields

    def get_rendered_fields(self) -> typing.List["fields.Field"]:
        return [field.render() for field in self.fields]

    def generate_extra_avro_attributes(self) -> None:
        """
        Look for the method in the dataclass.

        After calling the method extra_avro_attributes a dict is expected:
            typing.Dict[str, typing.Any]
        """
        extra_avro_attributes_fn = getattr(self.klass_or_instance, "extra_avro_attributes", None)

        if extra_avro_attributes_fn:
            extra_avro_attributes = extra_avro_attributes_fn()
            assert isinstance(extra_avro_attributes, dict), "Dict must be returned type in extra_avro_attributes method"

            aliases = extra_avro_attributes.get("aliases", self.aliases)
            namespace = extra_avro_attributes.get("namespace", self.namespace)

            self.aliases = aliases
            self.namespace = namespace

    def render(self):
        schema = OrderedDict(
            [("type", self.type), ("name", self.get_schema_name()), ("fields", self.get_rendered_fields()),]
        )

        if self.include_schema_doc:
            doc = self.generate_documentation()
            if doc is not None:
                schema["doc"] = doc

        if self.namespace is not None:
            schema["namespace"] = self.namespace

        if self.aliases is not None:
            schema["aliases"] = self.aliases

        return schema
